<copyright> It Document class.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of It.  It is distributed under the terms of the
    It License, a copy of which can be found in the It distribution; see
    the file LICENSE.

    <id>$Id: Document.t,v 1.21 1999/04/14 22:27:29 tiggr Exp $</id>
    </copyright>

// A Document should not be a subclass of View, but employ DocumentView
// objects for that purpose.  Document should be independent of the
// machinery used to manipulate its contents.
// Mon Mar 30 23:59:01 1998, tiggr@gerbil.org
implementation class
Document: View,
	it, XLTokens, tom.Constants
{
  <doc> The document currently being read.  </doc>
  public static Document current_document;
}

end;

implementation instance
Document
{
  <doc> The file of this document.  </doc>
  public String filename;

  <doc> The layers in this doc.  Upper layer first.  </doc>
  MutableArray layers;

  <doc> The layer getting newly created objects.  </doc>
  public mutable Layer active_layer;

  <doc> The tool getting mouse (dragged and up) events.  </doc>
  Tool active_tool;

  <doc> The selected objects.  These can reside in any of the layers.  For
      objects in the same layer, the index is increasing.  </doc>
  MutableArray selection;

  <doc> The modifications performed on this document.  </doc>
  MutableArray modifications;

  // Bit overkill?  perform afterDelay cancelPrevious?
  // Thu Apr 16 00:15:06 1998, tiggr@gerbil.org
  <doc> The timer scheduled to update the inspector.  </doc>
  Timer update_inspector;

  <doc> During a undo/redo sequence, this is one more than the index of
      the modification that has been performed last.  Modifications with
      an equal or higher index can be redone; modifications with a lower
      index can be undone.  </doc>
  int last_modification;

  <doc> Iff {TRUE}, we are printing.  Layers like the grid can use this to
      decide not to do anything.  </doc>
  public boolean printing;

  <doc> Iff {TRUE}, the selection update is not yet to be scheduled.  It
      will be scheduled when this becomes {FALSE}.  </doc>
  boolean inhibit_selection_update;
}

<doc> Designated initializer.  </doc>
id
      initFrame (float, float, float, float) (x, y, w, h)
  withFileNamed String fn
{
  modifications = [MutableObjectArray new];
  selection = [MutableObjectArray new];

  if (!fn)
    {
      MutableByteString buf = [MutableByteString new];
      int i;

      for (;;)
	{
	  [buf truncate 0];
	  [buf print ("untitled-", i++, ".it")];
	  if (![File file-exists buf])
	    break;
	}
      filename = buf;
    }
  else
    filename = fn;

  File f = [File open filename input: YES flags: FILE_NOT_EXIST_NOTHING];
  if (f != nil)
    {
      [self readFromStream f];
      [f close];
    }

  if (!layers)
    {
      if (!!f)
	[[[self error] print ("error reading file: ", filename)] nl];
      active_layer = [[ObjectLayer alloc] initWithDocument self];
      layers = [MutableObjectArray with active_layer];

      /* Now everything has been set up, we can notify.  */
      [[NotificationCenter default]
	postNotificationNamed notification_document_new object self];
    }
  else
    {
      [[NotificationCenter default]
	postNotificationNamed notification_document_open object self];

      [[[self diag] print ("read `", filename, "'")] nl];
    }

  = [super initFrame (x, y, w, h)];
}

<doc> Return {TRUE}.  </doc>
boolean
  acceptsFirstResponder MouseEvent event
{
  = TRUE;
}

<doc> Draw the ornaments of the objects in the {array}, which normally is
    the {selection}.  </doc>
void
	draw (float, float, float, float) (x, y, w, h)
  decoration Indexed array
{
  int i, n = [array length];

  for (i = 0; i < n; i++)
    {
      Object o = array[i];
      if ([tag rect [o decoratedBBox] intersects (x, y, w, h)])
	[o drawSelection (x, y, w, h) in self];
    }
}

<doc> Draw the objects intersecting the rectangle.  </doc>
void
     draw (float, float, float, float) (x, y, w, h)
  objects Indexed objects
{
  int i, n = [objects length];

  for (i = 0; i < n; i++)
    {
      Object o = objects[i];
      if ([tag rect (x, y, w, h) intersects [o decoratedBBox]])
	[o drawSelf (x, y, w, h) in self];
    }
}

<doc> Draw that part of this document that intersects with the given
    rectangle.  </doc>
void
  drawSelf (float, float, float, float) (x, y, w, h)
{
  tag.Display d = [self currentDisplay];

  if (!printing)
    {
      [d setColor [Color white]];
      [d fillRect (x, y, w, h)];
    }

  int i, n = [layers length];
  for (i = n - 1; i >= 0; i--)
    {
      Layer l = layers[i];
      [l drawSelf (x, y, w, h) in self];
    }

  if ([selection length] != 0)
    [self draw (x, y, w, h) decoration selection];

  if (active_tool)
    [active_tool drawActivity (x, y, w, h) in self];
}

/******************** files ********************/

protected void
  describe String description
   plugIts Keyed set
	to OutputStream s
{
  if ([set length] == 0)
    return;
  [s print ("(", description)];
  [set makeElementsPerform
       [Invocation of selector (void describePlugItRequirement OutputStream)
		   with s]];
  [[s print ')'] nl];
}

<doc> Read, as if we are loading, this document from the stream {s}.  </doc>
InputStream (s)
  readFromStream InputStream s
pre
  !current_document && !layers
post
  !!layers
{
  current_document = self;
  Lex lexer = [[Lex alloc] initWithStream s];
  Parse parser = [[Parse alloc] init Any (lexer)];
  layers = [parser it-file];
  current_document = nil;
}

<doc> Write, as if we are saving, this document to the stream {s}.  </doc>
OutputStream (s)
  writeToStream OutputStream s
{
  MutableSet display_set = [MutableEqSet new];
  MutableSet edit_set = [MutableEqSet new];

  [layers makeElementsPerform
	  [Invocation of selector (void collectPlugIts (MutableKeyed,
							MutableKeyed))
		      with (display_set, edit_set)]];

  [self describe "display-plugits" plugIts display_set to s];
  [self describe "edit-plugits" plugIts edit_set to s];

  [s print "(layers "];
  Encoder e = [[TextStreamEncoder alloc] initWithStream s];
  [e encodeRoot layers];
  [[s print ')'] nl];
}

/******************** layers ********************/

void
  addLayer Layer layer
  position int i
pre
  (i == LAYER_POSITION_BOTTOM
   || i == LAYER_POSITION_TOP
   || i < [layers length])
{
  if (i == LAYER_POSITION_BOTTOM)
    [layers add layer];
  else
    [layers insert layer at i];
}

/******************** objects ********************/

<doc> Insert a new modification record for the creation of the object {o}
    in the {modifications}, and redisplay after having made it the
    selection.  </doc>
void
  addObject Object object
{
  ObjectModification m = [ObjectModification new];
  [m setModifiedObject object at [active_layer addObject object]];
  [self addModification m];
  [self display [self setSelectedObject object]];
}

<doc> Add the modification {m} to {modifications} at the index
    {last_modification}, and truncate the {modifications} to its new
    length.  </doc>
void
  addModification Modification m
{
  modifications[last_modification++] = m;
  [modifications truncate last_modification];
}

<doc> Add the {object} to the selection, set its {selected}, and return
    the rectangle to be redrawn.  </doc>
(float, float, float, float)
  addToSelection Object object
pre
  ![object selected]
post
  [object selected]
{
  [object set_selected TRUE];
  = [object decoratedBBox];
  [self reflectSelection];

  int i, n = [selection length];
  ObjectLayer objlayer = [object layer];
  int objidx = [objlayer indexOfObject object];

  for (i = 0; i < n; i++)
    {
      Object o = selection[i];
      ObjectLayer l = [o layer];

      if (l == objlayer && [l indexOfObject o] > objidx)
	{
	  [selection insert object at i];
	  return;
	}
    }

  [selection add object];
}

<doc> Clear the current selection and redisplay.  </doc>
void
  clearSelection
{
  [self display [self setSelectedObject nil]];
}

<doc> Deselect the current selection and return the rectangle that needs
    to be redrawn.  </doc>
(float, float, float, float) (sx, sy, sw, sh)
  deselect
{
  int i, n = [selection length];

  if (n != 0)
    {
      for (i = 0; i < n; i++)
	{
	  Object o = selection[i];
	  (sx, sy, sw, sh) = [tag unionRect (sx, sy, sw, sh)
				  with [o decoratedBBox]];
	  [o set_selected FALSE];
	}
      [selection truncate 0];
      [self reflectSelection];
    }
}

<doc> Set the {inhibit_selection_update} flag.  If this becomes {FALSE},
    the {update_inspector} timer is scheduled if necessary.  </doc>
void
  inhibit_selection_update boolean inhibit_p
{
  if (!inhibit_p && update_inspector != nil && ![update_inspector scheduled])
    [update_inspector schedule];
  inhibit_selection_update = inhibit_p;
}

<doc> Schedule a timer to have us update the inspector, except when
    {inhibit_selection_update}, in which case we do create is but do not
    schedule it (that will be done later). </doc>
void
  reflectSelection
{
  if (!update_inspector)
    {
      update_inspector = [Timer withInterval 0.0 invocation
				[Invocation of selector (void updateSelection)
					    to: self]];
      if (!inhibit_selection_update)
	[update_inspector schedule];
    }
}

<doc> Invoked through a timer to update the inspector.  </doc>
void
  updateSelection
{
  update_inspector = nil;
  if (window == [[Application shared] main_window])
    [[InspectorManager shared] reflectSelection selection in self];
}

<doc> Remove the object {o} from the selection, unset its {selected}, and
    return the rectangle to be redrawn.  </doc>
(float, float, float, float)
  removeFromSelection Object o
pre
  [o selected]
post
  ![o selected]
{
  = [o decoratedBBox];
  [o set_selected FALSE];
  [selection removeElementAt [selection indexOfIdentical o]];
  [self reflectSelection];
}

<doc> Pass this method to the single selected object.  </doc>
(Shape, Point, Point) (shape, previous, next)
  selectedShapeSinglePoint
{
  if ([selection length] == 1)
    = [selection[0] selectedShapeSinglePoint];
}

<doc> Apply a rubberbanding selection with the given rectangle, and
    redisplay.  At least the second rectangle is redrawn.  </doc>
void
  selectRubber (float, float, float, float) (x, y, w, h)
	redraw (float, float, float, float) (rx, ry, rw, rh)
{
  int i, n = [layers length];

  for (i = 0; i < n; i++)
    {
      Layer l = layers[i];
      (rx, ry, rw, rh) = [tag unionRect (rx, ry, rw, rh)
			      with [l selectRubber (x, y, w, h)]];
    }

  [self display (rx, ry, rw, rh)];
  [self reflectSelection];
}

<doc> This returns the {selection}.  You are not only supposed to not
    modify it; also the objects contained in it should not be modified.
    </doc>
Indexed
  selection
{
  = selection;
}

<doc> Clear previous selection and make the {object} the selected object.
    Return the rectangle that needs to be redrawn to fix the changes to
    the display the document.  </doc>
(float, float, float, float) (sx, sy, sw, sh)
  setSelectedObject Object object
{
  (sx, sy, sw, sh) = [self deselect];
  if (object != nil)
    {
      [object set_selected TRUE];
      [selection add object];
      [self reflectSelection];
      = [tag unionRect (sx, sy, sw, sh) with [object decoratedBBox]];
    }
}

<doc> Clear previous selection and make the objects in the {array} the
    selected objects.  If {array} is {nil}, no objects will be selected.
    Return the rectangle that needs to be redrawn to fix the display of
    both the previous and current selection.  </doc>
(float, float, float, float) (sx, sy, sw, sh)
  setSelection Indexed array
{
  (sx, sy, sw, sh) = [self deselect];

  if (array != nil)
    {
      int i, n = [array length];

      if (n != 0)
	{
	  for (i = 0; i < n; i++)
	    {
	      Object o = array[i];

	      [o set_selected TRUE];
	      [selection add o];
	      (sx, sy, sw, sh) = [tag unionRect (sx, sy, sw, sh)
				      with [o decoratedBBox]];
	    }
	  [self reflectSelection];
	}
    }
}

/******************** events ********************/

<doc> Get the layers to identify the object being hit.  </doc>
(Object, int, All) (object, action, aux)
  hitTest (float, float) (x, y)
  partial boolean partial_p
{
  int i, priority, p, n = [layers length];
  float d, distance;
  Object o;
  All xtra;
  int a;

  for (i = 0; i < n; i++)
    {
      Layer l = layers[i];
      (o, a, xtra, d, p) = [l hitTest (x, y) partial partial_p];
      if (o != nil && (object == nil
		       || p > priority
		       || (p == priority && d < distance)))
	(object, action, aux, distance, priority) = (o, a, xtra, d, p);
    }
}

<doc> Let the layers hack the location of this event, if they so desire.
    </doc>
void
  lockEventLocation MouseEvent event
{
  float x, y;
  (x, y) = [self convertPoint [event location] fromView nil];

  int i, n = [layers length];
  float new_x, new_y;
  boolean locked;

  for (i = 0; i < n; i++)
    {
      (locked, new_x, new_y) = [layers[i] lockEvent event location (x, y)];
      if (locked)
	{
	  (x, y) = (new_x, new_y);
	  break;
	}
    }

  [event set_location (x, y)];
}

void
  mouseDown MouseEvent event
{
  Tool t = [[ToolKit shared] selected_tool];
  if (t != nil)
    {
      [self inhibit_selection_update TRUE];
      [self lockEventLocation event];
      if ([t mouseDown event in self])
	active_tool = t;
    }
}

void
  mouseDragged MouseEvent event
{
  if (active_tool != nil)
    {
      // Skip intermediate drag events...
      // Mon Apr 13 13:43:22 1998, tiggr@gerbil.org
      [self lockEventLocation event];
      [active_tool mouseDragged event in self];
    }
}

<doc> Send the {mouseUp in} to the {active_tool}, but not after having
    that particular variable to {nil}, so any redrawing before returning
    will not involve the effects of the tool.  </doc>
void
  mouseUp MouseEvent event
{
  if (active_tool != nil)
    {
      [self lockEventLocation event];
      Tool tool = active_tool;
      active_tool = nil;
      [tool mouseUp event in self];
      [self inhibit_selection_update FALSE];
    }
}

/******************** actions ********************/

void
  cut All sender
{
  [[self currentDisplay] beep];
}

void
  copy All sender
{
  [[self currentDisplay] beep];
}

void
  paste All sender
{
  [[self currentDisplay] beep];
}

void
  export All sender
{
  int i, n = [layers length];
  float sx, sy, sw, sh;

  printing = TRUE;
  for (i = 0; i < n; i++)
    {
      Layer l = layers[i];
      (sx, sy, sw, sh) = [tag unionRect (sx, sy, sw, sh)
			      with [l accurateBBox self]];
    }

  MutableArray save_selection = selection;
  selection = [MutableObjectArray withCapacity 0];

  MutableByteString exp_name
    = [[MutableByteString new]
	print ("ps:", [File basename filename without-extension: ".it"],
	       ".eps")];
  printer_display = [Display openQualifiedDisplay exp_name];
  if (!printer_display)
    return;

  [[[self diag] print ("exporting to `", exp_name, "'")] nl];

  unwind (
	  {
	    selection = save_selection;
	    [printer_display close];
	    printer_display = nil;
	    printing = FALSE;
	  })
    ({
      // Should have a printerdisplay class, so we do not need to emit
      // postscript code!
      // Thu Mar 12 23:03:36 1998, tiggr@gerbil.org
      [[[printer_display output]
	 print ("%%BoundingBox: ", sx, " ", sy, " ",
		sx + sw, " ", sy + sh)] nl];
      [self display];
    });
}

void
  save All sender
{
  OutputStream s;
  File f;

  if (!filename)
    s = [stdio out];
  else
    {
      f = [File open filename output: YES
		flags: FILE_EXIST_TRUNCATE | FILE_NOT_EXIST_CREATE];
      s = f;
    }

  [self writeToStream s];
  [[[self diag] print ("wrote `", filename, "'")] nl];

  if (!!f)
    [f close];
}

void
  undo All sender
{
  if (last_modification == 0)
    [[self currentDisplay] beep];
  else
    {
      Modification m = modifications[--last_modification];
      [self display [m undo self]];
    }
}

void
  redo All sender
{
  if (last_modification == [modifications length])
    [[self currentDisplay] beep];
  else
    {
      Modification m = modifications[last_modification++];
      [self display [m redo self]];
    }
}

end;
