<copyright> It Object class.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of It.  It is distributed under the terms of the
    It License, a copy of which can be found in the It distribution; see
    the file LICENSE.

    <id>$Id: Object.t,v 1.11 1998/04/19 14:59:53 tiggr Exp $</id>
    </copyright>

<doc> An object is the kind of thing that goes into a {Layer}.  </doc>
implementation class
Object: it, Copying

end;

implementation instance
Object
{
  <doc> The layer in which this object resides.  </doc>
  public mutable ObjectLayer layer;

  <doc> Bounding box of this object.  </doc>
  float b_x, b_y, b_w, b_h;

  <doc> Iff {TRUE}, this object is part of the selection.  </doc>
  public mutable boolean selected;
}

<doc> Designated initializer.  </doc>
id
  initWithBBox (float, float, float, float) (x, y, w, h)
{
  (b_x, b_y, b_w, b_h) = (x, y, w, h);
  = [super (it) init];
}

<doc> Return an accurate bounding box of the receiving object, suitable
    for use in for example exporting an EPS file.  The default
    implementation returns the {bbox}.  </doc>
(float, float, float, float)
  accurateBBox Document doc
{
  = [self bbox];
}

<doc> Given the {how} and {auxiliary} information that was returned by the
    receiving object as the object being hit for a {hitTest partial},
    adjust the selection of the document {doc}.  The selection is affected
    incrementally if {incremental_p}.

    If the receiving object wants to be moved by the manipulation through
    the selection tool, the {target} returned should be {nil}.  Otherwise,
    the {target} will receive repeated messages with the {action} selector
    during mouse-drag events, with the selection tool and the event as the
    two reference arguments (note: it is not a normal action).

    The default implementation sets the selection to the receiving object
    if not {incremental_p}; it toggles the selected state of the receiving
    object otherwise.  the {how} and {auxiliary} arguments are ignored,
    and the {target} that is returned will be {nil}.  </doc>
(All, selector) (target, action)
  adjustSelection (int, All) (how, auxiliary)
	       of Document doc
      incremental boolean incremental_p
{
  float sx, sy, sw, sh;

  if (incremental_p)
    if (!selected)
      (sx, sy, sw, sh) = [doc addToSelection self];
    else
      (sx, sy, sw, sh) = [doc removeFromSelection self];
  else if (!selected)
    (sx, sy, sw, sh) = [doc setSelectedObject self];
  else
    return;

  [doc display (sx, sy, sw, sh)];
}

<doc> Return the bounding box of the receiving object, i.e. {(b_x, b_y,
    b_w, b_h)}.  </doc>
(float, float, float, float)
  bbox
{
  = (b_x, b_y, b_w, b_h);
}

<doc> Add to the {display} and {edit} sets those plugits that are needed
    to display the receiving object, once retrieved from being saved, and
    the plugits needed to edit it, respectively.  The default
    implementation does nothing.  </doc>
void
  collectPlugIts (MutableKeyed, MutableKeyed) (display, edit)
{
}

<doc> Return an object that will be modified through the {target} and
    {action} of the {SelectionTool}.  The default implementation returns a
    deepened copy of {self}.  </doc>
id
  copyForModification
{
  = [[self copy] deepen -1];
}

<doc> Return the bounding box of the receiving object, including any
    decorations for its current selected state.  The returned box must
    contain the normal bbox.  The default implementation returns the
    {bbox}.  </doc>
(float, float, float, float)
  decoratedBBox
{
  = [self bbox];
}

<doc> Redraw the receiving object, as it is currently being modified.  The
    default implementation invokes {drawSelf} and, if {selected} also
    {drawSelection}.  </doc>
void
  drawModifyingSelf (float, float, float, float) (x, y, w, h)
		 in Document doc
{
  [self drawSelf (x, y, w, h) in doc];
  if (selected)
    [self drawSelection (x, y, w, h) in doc];
}

<doc> Draw the receiving object in the document {doc}.  The rectangle {(x,
    y, w, h)} intersects the bounding box of the receiving object.  </doc>
void
  drawSelf (float, float, float, float) (x, y, w, h)
	in Document doc
{
  Display d = [doc currentDisplay];

  [d setColor [Color black]];
  [d drawRect (b_x, b_y, b_w, b_h)];
  [d drawLine (b_x, b_y, b_x + b_w, b_y + b_h)];
  [d drawLine (b_x, b_y + b_h, b_x + b_w, b_y)];
}

<doc> Draw the ornaments of the receiving object, that indicate that the
    object has been selected.  </doc>
void
  drawSelection (float, float, float, float) (x, y, w, h)
	     in Document doc
{
}

<doc> For a mouse down event at {(x, y}), return the {object} being hit,
    some {action} implied by the particular part being hit for private
    use, the priority of the object (one of the {HIT_PRIO_*} constants),
    and the (relative) {distance} at which the object or a particular part
    of it is hit.  The default implementation returns a {nil} object.  In
    {auxiliary} any value may be returned that is important for the object
    later on.  </doc>
(Object, int, All, float, int) (object, action, auxiliary, distance, priority)
  hitTest (float, float) (x, y)
  partial boolean partial_p
{
}

<doc> Move the receiving object by {(dx, dy)}.  Or move it by {(x, y)}
    from its {original}.  The default implementation just moves the
    bounding box.  </doc>
void
  moveSelectedRelative (float, float) (dx, dy)
	    orAbsolute (float, float) (x, y)
		  from id original
{
  (b_x, b_y) = (b_x + dx, b_y + dy);
}

<doc> Return {TRUE} if only a partial overlap suffices for a rubberband
    selection to affect the receiving object.  The default implementation
    returns {FALSE}.  </doc>
boolean
  partialRubberSuffices
{
}

<doc> If the receiving object is a {Shape} with only the start or end
    point selected, return the shape and the start or end point (as {next}
    or {previous}, respectively).  The default implementation returns
    nothing.  </doc>
(Shape, Point, Point) (shape, previous, next)
  selectedShapeSinglePoint
post
  ((previous == nil || next == nil)
   && (!shape -> !previous && !next))
{
}

<doc> Apply rubber banding selection to the receiving object for the given
    rectangle.  The rectangle returned is the area of the document that
    needs to be redrawn to display the selection ornaments.

    The default implementation selects the receiving object if its bbox
    intersects or is included by the rubber rectangle (depending on
    {partialRubberSuffices}; the value returned is the decorated bbox
    (after selection).  </doc>
(float, float, float, float) (rx, ry, rw, rh)
  selectRubber (float, float, float, float) (x, y, w, h)
	    in Document doc
{
  if ([self partialRubberSuffices])
    {
      if (![tag rect (x, y, w, h) intersects [self bbox]])
	return;
    }
  else
    {
      float bx, by, bw, bh;
      (bx, by, bw, bh) = [self bbox];
      if (!([tag point (bx, by) inRect (x, y, w, h)]
	    && [tag point (bx + bw, by + bh) inRect (x, y, w, h)]))
	return;
    }

  if (!selected)
    = [doc addToSelection self];
}

/******************** Inspectors ********************/

<doc> Return the custom inspector for the object, in case it is the only
    selected object.  The default implementation returns the
    {PlainInspector}, with a text of {"No custom inspector"}.  </doc>
Inspector
  customInspector
{
  = [[PlainInspector shared] withTitle "No custom inspector"];
}

/******************** Archiving ********************/
<doc><h4>Archiving</h4></doc>

void
  encodeUsingCoder Encoder coder
{
  if (![coder hasBeenCodedFor [Object self]])
    {
      [super (it) encodeUsingCoder coder];
      [coder encode layer];
      [coder encode b_x];
      [coder encode b_y];
      [coder encode b_w];
      [coder encode b_h];
    }
}

void
  initWithCoder Decoder coder
{
  if (![coder hasBeenCodedFor [Object self]])
    {
      [super (it) initWithCoder coder];
      layer = [coder decode];
      b_x = [coder decode];
      b_y = [coder decode];
      b_w = [coder decode];
      b_h = [coder decode];
    }
}

end;
