<copyright> It ObjectLayer class.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of It.  It is distributed under the terms of the
    It License, a copy of which can be found in the It distribution; see
    the file LICENSE.

    <id>$Id: ObjectLayer.t,v 1.10 1998/04/13 16:55:30 tiggr Exp $</id>
    </copyright>

implementation class
ObjectLayer: Layer

end;

implementation instance
ObjectLayer
{
  <doc> The objects in this layer, the object at the back with the lowest
      index.  </doc>
  MutableArray objects;
}

<doc> Designated initializer.  </doc>
id
  initWithDocument Document doc
{
  objects = [MutableObjectArray new];
  = [super initWithDocument doc];
}

boolean
  acceptsNewObjects
{
  = TRUE;
}

<doc> Return the accurate bounding box of the contained objects.  </doc>
(float, float, float, float) (sx, sy, sw, sh)
  accurateBBox Document doc
{
  int i, n = [objects length];
  for (i = 0; i < n; i++)
    {
      Object o = objects[i];
      (sx, sy, sw, sh) = [tag unionRect (sx, sy, sw, sh)
			      with [o accurateBBox doc]];
    }
}

<doc> Add the {object} to the {objects}.  Return the index at which the
    object has been placed in the indexed collection of objects in this
    layer.  </doc>
int
  addObject Object object
{
  [object set_layer self];
  = [objects length];
  [objects add object];
}

<doc> Return the bounding box of the contained objects.  </doc>
(float, float, float, float) (sx, sy, sw, sh)
  bbox
{
  int i, n = [objects length];
  for (i = 0; i < n; i++)
    {
      Object o = objects[i];
      (sx, sy, sw, sh) = [tag unionRect (sx, sy, sw, sh) with [o bbox]];
    }
}

<doc> Pass this on to the {objects}.  </doc>
void
  collectPlugIts (MutableKeyed, MutableKeyed) (display, edit)
{
  [objects makeElementsPerform [Invocation of cmd with (display, edit)]];
}

<doc> Draw the objects intersecting the rectangle.  </doc>
void
  drawSelf (float, float, float, float) (x, y, w, h)
	in Document doc
{
  [doc draw (x, y, w, h) objects objects];
}

<doc> Return the object being hit (closest) within the receiving layer.
    </doc>
(Object, int, All, float, int) (object, action, auxiliary, distance, priority)
  hitTest (float, float) (x, y)
  partial boolean partial_p
{
  int i, p, n = [objects length];
  float d, bx, by, bw, bh;
  Object o;
  All xtra;
  int a;

  for (i = 0; i < n; i++)
    {
      Object c = objects[i];
      (bx, by, bw, bh) = [c decoratedBBox];
      if (x >= bx && x < bx + bw && y >= by && y < by + bh)
	{
	  (o, a, xtra, d, p) = [c hitTest (x, y) partial partial_p];
	  if (o != nil && (object == nil
			   || p > priority
			   || (p == priority && d < distance)))
	    (object, action, auxiliary, distance, priority)
	      = (o, a, xtra, d, p);
	}
    }
}

// Maybe the objects should know their index, since this is *slow*.
// Wed Mar 11 00:16:14 1998, tiggr@gerbil.org
<doc> Return the index of the {object} in the {objects}.  </doc>
int
  indexOfObject Object object
{
  = [objects indexOfIdentical object];
}

void
  insertObject Object o
	    at int i
post
  ([objects length] == old ([objects length]) + 1
   && objects[i] == o)
{
  [objects insert o at i];
}

void
  removeObject Object o
	    at int i
pre
  objects[i] == o
post
  [objects length] == old ([objects length]) - 1
{
  [objects removeElementAt i];
}

<doc> Select the points (and their objects) that lie within the given
    rectangle.  </doc>
(float, float, float, float) (rx, ry, rw, rh)
  selectRubber (float, float, float, float) (x, y, w, h)
{
  int i, n = [objects length];

  for (i = 0; i < n; i++)
    {
      Object o = objects[i];
      (rx, ry, rw, rh) = [tag unionRect (rx, ry, rw, rh)
			      with [o selectRubber (x, y, w, h) in document]];
    }
}

/******************** Archiving ********************/
<doc><h4>Archiving</h4></doc>

void
  encodeUsingCoder Encoder coder
{
  if (![coder hasBeenCodedFor [ObjectLayer self]])
    {
      [super encodeUsingCoder coder];
      [coder encode objects];
    }
}

void
  initWithCoder Decoder coder
{
  if (![coder hasBeenCodedFor [ObjectLayer self]])
    {
      [super initWithCoder coder];
      objects = [coder decode];
    }
}

id (self)
  awakeAfterUsingCoder Decoder coder
{
  // This should not be necessary...
  // Fri Apr 10 22:31:34 1998, tiggr@gerbil.org
  [[Document current_document] set_active_layer self];
  = [super awakeAfterUsingCoder coder];
}

end;
