<copyright> It ObjectModification class.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of It.  It is distributed under the terms of the
    It License, a copy of which can be found in the It distribution; see
    the file LICENSE.

    <id>$Id: ObjectModification.t,v 1.4 1998/03/11 00:05:16 tiggr Exp $</id>
    </copyright>

implementation class
ObjectModification: Modification

end;

implementation instance
ObjectModification
{
  <doc> The objects before the modification.  If this modification is one
      that created (direct or from a pastboard) new objects, this is
      {nil}.  </doc>
  Indexed original_objects;

  <doc> The index of each object in its layer (corresponding by index to
      the objects in the {original_objects}).  The indices referring to
      objects within the same layer must be in increasing order.  </doc>
  Indexed original_indices;

  <doc> The objects after modification.  If this modification removed
      objects without replacing them by something, this is {nil}.  </doc>
  Indexed modified_objects;

  <doc> And their indices.  The indices referring to objects within the
      same layer must be in increasing order.  </doc>
  Indexed modified_indices;
}

<doc> Designated initializer for modifications that create one or more
    objects to replace nothing.  </doc>
id
  init
{
  = [super init];
}

<doc> Designated initializer for modifications that replace {original}
    objects with the {modified} objects.  The {original_indices} and
    {modified_indices} are created from the information contained in the
    objects; the arrays {selection} and {modification} are used directly
    as the {orignal_objects} and {modified_objects}.  </doc>
id
  initWithObjects (Indexed, Indexed) (selection, modification)
pre
  [selection length] == [modification length]
{
  int i, n = [selection length];
  MutableIntArray indices = [MutableIntArray withCapacity n];

  original_indices = modified_indices = indices;
  original_objects = selection;
  modified_objects = modification;

  for (i = 0; i < n; i++)
    {
      Object sel = selection[i];
      Object mod = modification[i];
      int idx = [[sel layer] indexOfObject sel];
      [indices add idx];
    }

  = [super init];
}

<doc> Un/re/do part of this modification by adding the {objects} to their
    layers.  </doc>
protected void
  add (Indexed, Indexed) (objects, indices)
pre
  [objects length] == [indices length]
{
  int i, n = [objects length];

  for (i = 0; i < n; i++)
    {
      Object o = objects[i];
      [[o layer] insertObject o at indices[i]];
    }
}

<doc> Un/re/do part of this modification by removing the {objects} from
    their layers.  </doc>
protected void
  remove (Indexed, Indexed) (objects, indices)
pre
  [objects length] == [indices length]
{
  int i, n = [objects length];

  for (i = n - 1; i >= 0; i--)
    {
      Object o = objects[i];
      [[o layer] removeObject o at indices[i]];
    }
}

<doc> Set the single post-object.  </doc>
void
  setModifiedObject Object object
		 at int index
pre
  !original_objects || [original_objects length] == 1
{
  modified_objects = [ObjectArray with object];
  modified_indices = [IntArray with index];
}

<doc> Redo the effect of this modification.  </doc>
(float, float, float, float) (sx, sy, sw, sh)
  redo Document doc
{
  if (original_objects != nil)
    {
      (sx, sy, sw, sh) = [doc setSelection original_objects];
      [self remove (original_objects, original_indices)];
    }

  if (modified_objects != nil)
    [self add (modified_objects, modified_indices)];

  = [tag unionRect (sx, sy, sw, sh)
	 with [doc setSelection modified_objects]];
}

<doc> Undo the effect of this modification to the Document {doc}.  </doc>
(float, float, float, float) (sx, sy, sw, sh)
  undo Document doc
{
  if (modified_objects != nil)
    {
      (sx, sy, sw, sh) = [doc setSelection modified_objects];
      [self remove (modified_objects, modified_indices)];
    }

  if (original_objects != nil)
    [self add (original_objects, original_indices)];

  = [tag unionRect (sx, sy, sw, sh)
	 with [doc setSelection original_objects]];
}

<doc> Emit the objects and indices.  </doc>
OutputStream (s)
  writeFields OutputStream s
{
  = [super writeFields s];
  = [s print (" orig=", original_objects)];
  = [s print (" oidx=", original_indices)];
  = [s print (" mods=", modified_objects)];
  = [s print (" midx=", modified_indices)];
}

end;
