<copyright> It PlugIt class.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of It.  It is distributed under the terms of the
    It License, a copy of which can be found in the It distribution; see
    the file LICENSE.

    <id>$Id: PlugIt.t,v 1.4 1998/03/22 21:18:18 tiggr Exp $</id>
    </copyright>

<doc> A PlugIt is an It plug-in.  </doc>
implementation class
PlugIt: it, tom.Constants
{
  <doc> All registered plug-its.  </doc>
  static MutableMapped all_plug_its;

  <doc> The shared plugit of this particular plugit class.  </doc>
  instance (id) shared;
}

<doc> This method is invoked when a plug-it has just been loaded.  The
    default implementation allocates a new instance and passes it to
    {registerPlugIt}.  Override to provide specialized behaviour.

    This method exists to prevent every plug-it from having a {load}
    method for just this purpose.  </doc>
void
  initializePlugIt
{
  [self registerPlugIt [self shared]];
}

// Use Bundle to load; drop the second argument.
// Sun Feb 22 21:47:54 1998, tiggr@gerbil.org
void
  loadPlugIt String name
  fromObject String filename
{
  pointer success = [Bundle load filename];
  if (!!success)
    {
      MutableByteString s = [[MutableByteString new] print (name, '.', name)];
      id cls = [Runtime classNamed s];

      if (!cls)
	[[[self error] print ("No class `", s, "' in object ", filename)] nl];
      else
	[cls initializePlugIt];
    }
}

<doc> Load all plug-its from the application directory.  </doc>
void
  loadPlugIts
{
  String dir = [[Bundle main] directory];
  Indexed files = [File filenames-in-directory dir];
  MutableByteString buf = [MutableByteString withCapacity 80];
  int i, n = [files length];

  for (i = 0; i < n; i++)
    {
      String file = files[i];

      [buf truncate 0];
      [buf print (dir, file)];
      if ([File type-of-file buf] == FILE_TYPE_DIRECTORY)
	{
	  [buf print ("/", file, ".o")];
	  if ([File file-exists buf])
	    [self loadPlugIt file fromObject buf];
	}
    }
}

<doc> Return the plugit with the {name}, or {nil} if it is currently not
    loaded.  </doc>
instance (id)
  named String name
{
  = all_plug_its[name];
}

<doc> Return a readable string for a version number.  </doc>
String
  readableVersion int version
{
  = [[MutableByteString new] print (version / 1000000, ".",
				    (version / 1000) % 1000, ".",
				    version % 1000)];
}

<doc> Register the plug-it {p}.  </doc>
protected void
  registerPlugIt PlugIt p
{
  if (!all_plug_its)
    all_plug_its = [MutableDictionary new];

  String n = [p plugItName];

  [[[self diag] print ("loaded plug-it ", n, ", version ",
		       [self readableVersion [p plugItVersion]])] nl];
  all_plug_its[n] = p;
}

<doc> Return the plugit object of the receiving plugit class.  </doc>
instance (id)
  shared
{
  if (!shared)
    shared = [self new];
  = shared;
}

end;

implementation instance
PlugIt

<doc> Describe the dependence on the receiving plugit to the {stream}.
    </doc>
void
  describePlugItRequirement OutputStream stream
{
  [stream print (" (check-plugit \"", [self plugItName], "\" ",
		 [self plugItVersionLast], ")")];
}

String
  plugItName
{
  = "unnamed plug-it";
}

<doc> Return the version number of this plugit.  This resembles xxxyyyzzz,
    where xxx is the major version number, yyy is the minor version number
    (3 digits), and zzz is a patchlevel (3 digits).  The default version
    is 0, suitable for plugits in early development.  </doc>
int
  plugItVersion
{
}

<doc> Return the oldest plugit version that is compatible, at the document
    file level, with the current plugit.  The default implementation
    truncates the patchlevel and minor version.  </doc>
int
  plugItVersionLast
{
  = 1000000 * ([self plugItVersion] / 1000000);
}

end;
