<copyright> It Point class.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of It.  It is distributed under the terms of the
    It License, a copy of which can be found in the It distribution; see
    the file LICENSE.

    <id>$Id: Point.t,v 1.9 1998/04/19 14:59:54 tiggr Exp $</id>
    </copyright>

implementation class
Point: it, Copying, Math
{
  <doc> The dimensions of the {point_image}.  </doc>
  static float pi_w, pi_h;

  <doc> The image for displaying a selected point.  </doc>
  static Image point_selected_image;

  <doc> The dimensions of the {point_selected_image}.  </doc>
  static float psi_w, psi_h;

  <doc> The image for displaying a control point.  </doc>
  static Image point_control_image;

  <doc> And its dimensions.  </doc>
  static float pci_w, pci_h;

  <doc> Kinds of points.  </doc>
  const KIND_CORNER = 0;
  const KIND_CURVE = 2;
}

<doc> Make sure the images have been fetched before passing to {super}.
    </doc>
instance (id)
  alloc
{
  if (!point_selected_image)
    {
      (pi_w, pi_h) = [point_image size];
      point_selected_image = [Image named "point-selected"];
      (psi_w, psi_h) = [point_selected_image size];
      point_control_image = [Image named "point-control"];
      (pci_w, pci_h) = [point_control_image size];
    }

  = [super (class (it)) alloc];
}

end;

implementation instance
Point
{
  <doc> The position of the control point of the curve finishing at this
      point.  </doc>
  float to_x, to_y;

  <doc> The position of this point.  </doc>
  float pos_x, pos_y;

  <doc> The position of the control point of the curve starting from this
      point.  </doc>
  float from_x, from_y;

  <doc> The kind of point this is.  </doc>
  int kind;

  <doc> Iff {TRUE}, this point is selected.  </doc>
  public mutable boolean selected;
}

<doc> Designated initializer.  </doc>
id
      init Object containing_object
  position (float, float) (x, y)
{
  (pos_x, pos_y) = (x, y);
  = [super (it) init];
}

<doc> Return the bbox of this point, without any decorations.  </doc>
(float, float, float, float) (x, y, w, h)
  bbox
{
  (x, y, w, h) = (pos_x, pos_y, 1.0, 1.0);
  if (kind == KIND_CURVE)
    (x, y, w, h) = [tag unionRect [tag unionRect (x, y, w, h)
				       with (to_x, to_y, 1.0, 1.0)]
			with (from_x, from_y, 1.0, 1.0)];
}

<doc> Return the bbox of the decoration of the incoming or outgoing curve
    control point.  </doc>
(float, float, float, float) (x, y, w, h)
  controlDecorationBBox boolean incoming_not_outgoing
{
  (w, h) = selected ? (psi_w, psi_h) : (pi_w, pi_h);
  (x, y) = (pos_x - w / 2.0, pos_y - h / 2.0);

  // A curve incoming to a non-curve also has an incoming control point.
  // Mon Apr 13 00:09:23 1998, tiggr@gerbil.org
  if (kind == KIND_CURVE)
    if (incoming_not_outgoing)
      (x, y, w, h) = [tag unionRect (x, y, w, h)
			  with (to_x - pci_w / 2.0, to_y - pci_h / 2.0,
				pci_w, pci_h)];
    else
      (x, y, w, h) = [tag unionRect (x, y, w, h)
			  with (from_x - pci_w / 2.0, from_y - pci_h / 2.0,
				pci_w, pci_h)];
}

<doc> Return the bbox of this point, using the apperance that is suitable
    for its {selected} state, and the state of the {previous} and {next}
    points.  </doc>
(float, float, float, float) (x, y, w, h)
  decoratedBBox (id, id) (previous, next)
{
  (w, h) = selected ? (psi_w, psi_h) : (pi_w, pi_h);
  (x, y) = (pos_x - w / 2.0 + 0.5, pos_y - h / 2.0 + 0.5);

  if (kind == KIND_CURVE)
    {
      /* We could discern being selected or not, but then there are too
         many paths through code where that information has to be
         maintained or propagated.  */
      (x, y, w, h)
	= [tag unionRect [tag unionRect (x, y, w, h)
			      with (to_x - pci_w / 2.0, to_y - pci_h / 2.0,
				    pci_w, pci_h)]
	       with (from_x - pci_w / 2.0, from_y - pci_h / 2.0,
		     pci_w, pci_h)];
    }
}

<doc> Return the decorated bbox of this point, including the bbox of any
    decoration of the neighbours because the receiving point is selected.
    </doc>
(float, float, float, float) (x, y, w, h)
  decorationBBox (id, id) (previous, next)
{
  (x, y, w, h) = [self decoratedBBox (previous, next)];
  if (previous != nil)
    (x, y, w, h) = [tag unionRect (x, y, w, h)
			with [previous controlDecorationBBox FALSE]];
  if (next != nil)
    (x, y, w, h) = [tag unionRect (x, y, w, h)
			with [next controlDecorationBBox TRUE]];
}

<doc> Draw the decoration of the receiving point.  </doc>
void
  drawSelection (float, float, float, float) (x, y, w, h)
	     in Document doc
	   with (id, id) (previous, next)
{
  float rx, ry, rw, rh;
  (rx, ry, rw, rh) = [self decoratedBBox (previous, next)];

  if ([tag rect (rx, ry, rw, rh) intersects (x, y, w, h)])
    {
      boolean sel_p = previous != nil && [previous selected];
      boolean sel_n = next != nil && [next selected];
      Display d = [doc currentDisplay];

      if (kind == KIND_CURVE)
	{
	  [d setColor [Color black]];
	  if (selected || sel_p)
	    [d drawLine (pos_x, pos_y, to_x, to_y)];
	  if (selected || sel_n)
	    [d drawLine (pos_x, pos_y, from_x, from_y)];
	}

      [isa centerImage (selected ? point_selected_image : point_image)
	   at (pos_x, pos_y)];

      if (kind == KIND_CURVE)
	{
	  if (selected || sel_p)
	    [isa centerImage point_control_image at (to_x, to_y)];
	  if (selected || sel_n)
	    [isa centerImage point_control_image at (from_x, from_y)];
	}
    }
}

<doc> Draw the receiving point.  </doc>
void
  drawSelf Display d
      with (id, id) (previous, next)
{
  if (!previous)
    [d moveto (pos_x, pos_y)];
  else if (kind == KIND_CORNER)
    [d lineto (pos_x, pos_y)];
  else
    {
      float x1, y1;
      (x1, y1) = [previous locationOutgoing];
      [d curveto (x1, y1, to_x, to_y, pos_x, pos_y)];
    }
}

(boolean, int, float, int) (hit_p, action, distance, priority)
  hitTest (float, float) (x, y)
     with (id, id) (previous, next)
{
  float bx, by, bw, bh;

  // Is this fast?
  // Mon Apr 13 01:43:20 1998, tiggr@gerbil.org
  (bx, by, bw, bh) = [self decoratedBBox (previous, next)];
  if (x >= bx && x < bx + bw && y >= by && y < by + bh)
    {
      float dt, df, min;
      float dp = min = (x - pos_x) * (x - pos_x) + (y - pos_y) * (y - pos_y);
      action = SHAPE_POINT_POSITION;

      if (selected || (previous != nil && [previous selected]))
	{
	  dt = (x - to_x) * (x - to_x) + (y - to_y) * (y - to_y);
	  if (dt < min)
	    {
	      action = SHAPE_POINT_TO;
	      min = dt;
	    }
	}

      if (selected || (next != nil && [next selected]))
	{
	  df = (x - from_x) * (x - from_x) + (y - from_y) * (y - from_y);
	  if (df < min)
	    {
	      action = SHAPE_POINT_FROM;
	      min = df;
	    }
	}

      distance = float (sqrt (min));
      // Arbitrary bound...
      // Mon Apr 13 01:50:43 1998, tiggr@gerbil.org
      if (distance < 3.0)
	(hit_p, priority) = (YES, (action == SHAPE_POINT_POSITION
				   ? HIT_PRIO_DETAIL : HIT_PRIO_DECORATION));
    }
}

<doc> Return {(pos_x, pos_y)}.  </doc>
(float, float)
  location
{
  = (pos_x, pos_y);
}

<doc> Return the location of the incoming control point,
    i.e. {(to_x, to_y)}.  </doc>
(float, float)
  locationIncoming
{
  if (kind != KIND_CURVE)
    [self unimplemented cmd message: "no curve"];

  = (to_x, to_y);
}

<doc> Return the location of the outgoing control point,
    {(from_x, from_y)}.  </doc>
(float, float)
  locationOutgoing
{
  if (kind != KIND_CURVE)
    [self unimplemented cmd message: "no curve"];

  = (from_x, from_y);
}

<doc> Move one of the control points, indicated by {how}, a distance {(dx,
    dy)} from the {original} point.  Rotate the other control point around
    the location of this point, to ensure the curve remains continuous.
    </doc>
void
	  move (float, float) (dx, dy)
  controlPoint int how
	  from id original
// Gi does not grok this, and the compiler does not grok
// (pos_x, pos_y) == [original location]
// Mon Apr 13 15:13:27 1998, tiggr@gerbil.org
//pre
//  (({float x, y; (x, y) = [original location]; pos_x == x && pos_y == y;})
//   && (how == SHAPE_POINT_TO || how == SHAPE_POINT_FROM))
{
  float odx, ody, inx, iny, outx, outy;

  (inx, iny) = [original locationIncoming];
  (outx, outy) = [original locationOutgoing];

  if (how == SHAPE_POINT_TO)
    {
      (to_x, to_y) = (inx + dx, iny + dy);
      (dx, dy) = (to_x - pos_x, to_y - pos_y);
      (odx, ody) = (from_x - pos_x, from_y - pos_y);
    }
  else
    {
      (from_x, from_y) = (outx + dx, outy + dy);
      (dx, dy) = (from_x - pos_x, from_y - pos_y);
      (odx, ody) = (to_x - pos_x, to_y - pos_y);
    }

  /* (DX, DY) is now the vector from the position of this point to the
     control point being moved by the mouse.  Put the other control point
     on the same line.  First normalize (DX, DY).  */
  float norm1 = float (sqrt (dx * dx + dy * dy));
  (dx, dy) = (dx / norm1, dy / norm1);

  /* NORM2 is the length of the other control point.  */
  float norm2 = float (sqrt (odx * odx + ody * ody));

  if (how == SHAPE_POINT_TO)
    (from_x, from_y) = (pos_x - dx * norm2, pos_y - dy * norm2);
  else
    (to_x, to_y) = (pos_x - dx * norm2, pos_y - dy * norm2);
}

<doc> Move the point.  </doc>
void
  moveBy (float, float) (dx, dy)
{
  (pos_x, pos_y) = (pos_x + dx, pos_y + dy);
  if (kind == KIND_CURVE)
    {
      (from_x, from_y) = (from_x + dx, from_y + dy);
      (to_x, to_y) = (to_x + dx, to_y + dy);
    }
}

void
  set_control_points (float, float, float, float) (incoming_x, incoming_y,
						   outgoing_x, outgoing_y)
{
  (to_x, to_y) = (incoming_x, incoming_y);
  (from_x, from_y) = (outgoing_x, outgoing_y);
}

void
  set_is_curve
{
  kind = KIND_CURVE;
}

<doc> Emit the location.  </doc>
OutputStream (s)
  writeFields OutputStream s
{
  = [[super (it) writeFields s] print (" (", pos_x, ", ", pos_y, ")")];
  if (kind == KIND_CURVE)
    {
      = [s print (" from=(", from_x, ", ", from_y, ")")];
      = [s print (" to=(", to_x, ", ", to_y, ")")];
    }    
}

/******************** Archiving ********************/
<doc><h4>Archiving</h4></doc>

void
  encodeUsingCoder Encoder coder
{
  if (![coder hasBeenCodedFor [Point self]])
    {
      [super (it) encodeUsingCoder coder];
      [coder encode pos_x];
      [coder encode pos_y];
      [coder encode kind];
      if (kind == KIND_CURVE)
	{
	  [coder encode to_x];
	  [coder encode to_y];
	  [coder encode from_x];
	  [coder encode from_y];
	}
    }
}

void
  initWithCoder Decoder coder
{
  if (![coder hasBeenCodedFor [Point self]])
    {
      [super (it) initWithCoder coder];
      pos_x = [coder decode];
      pos_y = [coder decode];
      kind = [coder decode];
      if (kind == KIND_CURVE)
	{
	  to_x = [coder decode];
	  to_y = [coder decode];
	  from_x = [coder decode];
	  from_y = [coder decode];
	}
    }
}

end;
