<copyright> It SelectionTool class.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of It.  It is distributed under the terms of the
    It License, a copy of which can be found in the It distribution; see
    the file LICENSE.

    <id>$Id: SelectionTool.t,v 1.11 1998/04/19 14:59:55 tiggr Exp $</id>
    </copyright>

implementation class
SelectionTool: Tool

<doc> Return {"tool-select"} and {"tool-select-h"}.  </doc>
(String, String) (normal, alternate)
  cell_images
{
  = ("tool-select", "tool-select-h");
}

end;

implementation instance
SelectionTool
{
  <doc> The most recent mouse down location, in the coordinate system of
      the document.  </doc>
  float down_x, down_y;

  <doc> Likewise, the most recent mouse drag location.  </doc>
  float drag_x, drag_y;

  <doc> The object that is to receive the {action} messages in response to
      mouse drags.  If this is not set, mouse drags will move the
      {selection} (or parts thereof, as is done by {Shape}s with
      explicitly selected {Point}s.  </doc>
  All target;

  <doc> The selector of the message to be sent to the {target}.  This is
      only valid if the {target} is not {nil}.  Unlike normal action
      selectors (which accept one reference, the sender), this one must
      accept 5 references: the {target_original} and {target_replacement},
      the selection tool (us), the document, and the (mouse drag) event.
      </doc>
  selector action;

  <doc> When dragging through a {target}, this is the original object
      being modified.  </doc>
  Object target_original;

  <doc> Likewise, this is the replacement object.  It is this object that
      is drawn from the {drawActivity in} of this selection tool.  </doc>
  Object target_replacement;

  <doc> When dragging a bunch of objects, this array contains the original
      objects making up the selection being dragged.  </doc>
  Indexed selection;

  <doc> When dragging objects, these are the objects in their new
      position, orientation, scale, whatever.  </doc>
  Indexed modification;

  <doc> While dragging objects, this is the bounding box of the
      {modification} objects.  </doc>
  float msx, msy, msw, msh;

  <doc> Iff {TRUE}, we are rubber banding the selection.  Only valid
      during {mouseDragged} and {mouseUp}.  </doc>
  boolean rubber;
}

<doc> Return the location of the (most recent) mouse down.  </doc>
(float, float)
  mouseDownLocation
{
  = (down_x, down_y);
}

<doc> Return the location of the (most recent) mouse drag.  </doc>
(float, float)
  mouseDragLocation
{
  = (drag_x, drag_y);
}

<doc> Draw the rubber banding.  </doc>
void
  drawActivity (float, float, float, float) (x, y, w, h)
	    in Document doc
{
  Display d = [doc currentDisplay];

  if (rubber)
    {
      float r_w = drag_x - down_x, r_h = drag_y - down_y;
      float pr_x, pr_y, pr_w, pr_h;

      (pr_x, pr_y, pr_w, pr_h) = fix_rectangle (down_x, down_y, r_w, r_h);
      (pr_x, pr_y, pr_w, pr_h)
	= (pr_x - 1.0, pr_y - 1.0, pr_w + 3.0, pr_h + 3.0);

      if ([tag rect (pr_x, pr_y, pr_w, pr_h) intersects (x, y, w, h)])
	{
	  [d setColor [Color white]];
	  [d drawRect (pr_x, pr_y, pr_w, pr_h)];
	  [d drawRect (pr_x + 2.0, pr_y + 2.0, pr_w - 4.0, pr_h - 4.0)];
	  [d setColor [Color black]];
	  [d drawRect (pr_x + 1.0, pr_y + 1.0, pr_w - 2.0, pr_h - 2.0)];
	}
    }
  else if (!target)
    {
      [doc draw (x, y, w, h) objects modification];
      // Tue Mar 10 23:48:34 1998, tiggr@gerbil.org
      // [doc draw (x, y, w, h) decoration modification];
    }
  else if (target_replacement != nil
	   && [tag rect (x, y, w, h)
		   intersects [target_replacement decoratedBBox]])
    [target_replacement drawModifyingSelf (x, y, w, h) in doc];
}

boolean
  mouseDown MouseEvent e
	 in Document doc
{
  int how;
  All aux;

  (down_x, down_y) = (drag_x, drag_y) = [e location];
  int mods = [e modifiers];

  // Unconfigurable MOD_SHIFT is acceptable; Likewise MOD_CONTROL is not.
  // Mon Apr 13 01:56:52 1998, tiggr@gerbil.org
  boolean partial = !!(mods & EVENT_MOD_CONTROL);
  (target_original, how, aux) = [doc hitTest (down_x, down_y) partial partial];

  boolean incremental = !!(mods & EVENT_MOD_SHIFT);
  if (!target_original)
    {
      /* We'll do rubber banding.  */
      // Incrementally...
      // Mon Apr 13 00:47:25 1998, tiggr@gerbil.org
      [doc clearSelection];
      rubber = TRUE;
    }
  else
    {
      rubber = FALSE;
      (target, action) = [target_original adjustSelection (how, aux) of doc
					  incremental incremental];
      if (!target)
	{
	  /* If the selection ends up empty, there's not much we need to
             do.  */
	  Indexed docsel = [doc selection];
	  if (![docsel length])
	    return FALSE;
	}
      else
	target_replacement = nil;
    }

  = TRUE;
}

void
  mouseDragged MouseEvent e
	    in Document doc
{
  float x, y;
  (x, y) = [e location];
  if (x == drag_x && y == drag_y)
    return;

  if (rubber)
    {
      float r_w = drag_x - down_x, r_h = drag_y - down_y;

      /* Previous selection rectangle bbox.  */
      float pr_x, pr_y, pr_w, pr_h;
      (pr_x, pr_y, pr_w, pr_h) = fix_rectangle (down_x, down_y, r_w, r_h);
      (pr_x, pr_y, pr_w, pr_h)
	= (pr_x - 1.0, pr_y - 1.0, pr_w + 3.0, pr_h + 3.0);

      (drag_x, drag_y) = (x, y);
      (r_w, r_h) = (drag_x - down_x, drag_y - down_y);

      /* Current selection rectangle bbox.  */
      float nr_x, nr_y, nr_w, nr_h;
      (nr_x, nr_y, nr_w, nr_h) = fix_rectangle (down_x, down_y, r_w, r_h);
      (nr_x, nr_y, nr_w, nr_h)
	= (nr_x - 1.0, nr_y - 1.0, nr_w + 3.0, nr_h + 3.0);

      // Tad bit of a coarse redraw...
      // Fri Feb 27 19:21:44 1998, tiggr@gerbil.org
      [doc display [tag unionRect (pr_x, pr_y, pr_w, pr_h)
			with (nr_x, nr_y, nr_w, nr_h)]];
      return;
    }

  /* Are we here for the first time?  */
  if (!selection)
    {
      if (target != nil)
	selection = [[doc selection] copy];
      else
	{
	  selection = [[doc selection] copy];
	  modification = [[selection copy] deepen 2];
	}
      (msx, msy, msw, msh) = [doc setSelectedObject nil];

      // This is needed, for now, to remove the decoration of the
      // selection, in case TARGET != nil.
      // Mon Apr 13 16:03:05 1998, tiggr@gerbil.org
      [doc display (msx, msy, msw, msh)];
    }

  if (!target)
    {
      /* We are moving the selection.  */
      float dx, dy, nsx, nsy, nsw, nsh;
      (dx, dy) = (x - drag_x, y - drag_y);
      (drag_x, drag_y) = (x, y);

      int i, n = [selection length];
      for (i = 0; i < n; i++)
	{
	  Object modified = modification[i];

	  [modified moveSelectedRelative (dx, dy)
		    orAbsolute (drag_x - down_x, drag_y - down_y)
		    from selection[i]];
	  (nsx, nsy, nsw, nsh) = [tag unionRect (nsx, nsy, nsw, nsh)
				      with [modified decoratedBBox]];
	}

      [doc display [tag unionRect (msx, msy, msw, msh)
			with (nsx, nsy, nsw, nsh)]];
      (msx, msy, msw, msh) = (nsx, nsy, nsw, nsh);
    }
  else
    {
      /* The target is modifying itself.  */
      if (!target_replacement)
	target_replacement = [target_original copyForModification];

      (drag_x, drag_y) = (x, y);
      [target perform action with (target_original, target_replacement,
				   doc, self, e)];
    }
}

void
  mouseUp MouseEvent e
       in Document doc
{
  if (rubber)
    {
      if (drag_x != down_x || drag_y != down_y)
	{
	  /* If you witness artefacts after rubber banding, the assumpting
             that the mouse up will not introduce a new position is wrong.  */
	  float r_w = drag_x - down_x, r_h = drag_y - down_y;

	  float pr_x, pr_y, pr_w, pr_h;
	  (pr_x, pr_y, pr_w, pr_h) = fix_rectangle (down_x, down_y, r_w, r_h);

	  [doc selectRubber (pr_x, pr_y, pr_w + 1.0, pr_h + 1.0)
	       redraw (pr_x - 2.0, pr_y - 2.0, pr_w + 4.0, pr_h + 4.0)];
	}
      return;
    }

  if (drag_x == down_x && drag_y == down_y)
    {
      /* We haven't moved an inch.  Possibly because we didn't move at
	 all, or because we returned to the original position.  */
      if (selection != nil)
	[doc display [doc setSelection selection]];
      else
	{
	  // Too coarse...
	  // Mon Apr 13 15:42:32 1998, tiggr@gerbil.org
	  [doc display];
	}
    }
  else
    {
      if (target != nil)
	{
	  selection = [ObjectArray with target_original];
	  modification = [ObjectArray with target_replacement];
	}

      ObjectModification m = [[ObjectModification alloc]
			       initWithObjects (selection, modification)];
      [doc addModification m];
      [doc display [m redo doc]];
    }

  selection = modification = nil;
  target = target_original = target_replacement = nil;
}

end;
