<copyright> It Shape class.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of It.  It is distributed under the terms of the
    It License, a copy of which can be found in the It distribution; see
    the file LICENSE.

    <id>$Id: Shape.t,v 1.12 1999/04/14 22:27:33 tiggr Exp $</id>
    </copyright>

<doc> A {Shape} is an {Object} which employs some {Points}.  </doc>
implementation class
Shape: Object

<doc> Return the current archiving version.  </doc>
int
  version
{
  if (self == [Shape self])
    = 1;
}

end;

implementation instance
Shape
{
  <doc> The points of this shape.  </doc>
  MutableArray points;

  <doc> The number of selected points in this shape.  </doc>
  int num_selected;

  <doc> Iff {TRUE}, this path is closed.  </doc>
  boolean closed;
}

<doc> Designated initializer.  The {points} must have been filled in by
    the subclass.  </doc>
id
  initWithBBox (float, float, float, float) (x, y, w, h)
pre
  points != nil
{
  = [super initWithBBox (x, y, w, h)];
}

<doc> Designated initializer to start a shape with a single point.  </doc>
id
  initWithPoint Point point
{
  points = [MutableObjectArray with point];
  float x, y;
  (x, y) = [point location];
  if ([point selected])
    num_selected = 1;
  = [super initWithBBox (x, y, 1.0, 1.0)];
}

<doc> Return the bbox of the path after flattening.  </doc>
(float, float, float, float) (x, y, w, h)
  accurateBBox Document doc
{
  [doc lockFocus];
  Display d = [doc currentDisplay];
  [self setupCurrentPath d];

  [d flattenpath];
  float x1, y1, x2, y2;
  (x1, y1, x2, y2) = [d pathbbox];
  [doc unlockFocus];

  = (x1 - 0.5, y1 - 0.5, x2 - x1 + 0.5, y2 - y1 + 0.5);
}

<doc> Add the {point} to the receiving shape, at index {i}, and recompute
    the bbox.  Return the index at which it was inserted.  </doc>
int (index)
  addPoint Point point
  position int i
pre
  SHAPE_POINT_LAST <= i && i < [points length]
post
  (0 <= index && index < [points length]
   && [points length] == old ([points length]) + 1)
{
  if (i == SHAPE_POINT_LAST)
    {
      = [points length];
      points[index] = point;
    }
  else
    {
      [points insert point at i];
      = i;
    }

  if ([point selected])
    if (selected)
      num_selected++;
    else
      [point set_selected FALSE];

  [self recomputeBBox];
}

<doc> Adjust the selected status of the {point}.  If it is {nil} however,
    or when the receiving shape is currently unselected, let {super} do
    its work.  If the {how} is one of the control-point motion actions,
    the selection is not adjusted at all, and the {target} returned will
    be a curried {Invocation} to {self}, for one of the point-moving
    methods.  </doc>
(All, selector) (target, action)
  adjustSelection (int, Point) (how, point)
	       of Document doc
      incremental boolean incremental_p
pre
  !point || [points indexOfIdentical point] >= 0
{
  if (how == SHAPE_POINT_TO || how == SHAPE_POINT_FROM)
    return ([Invocation for selector
				(void moveControlPoint (int, Point)
				      of (id, id) in Document
				      by (SelectionTool, MouseEvent))
			to: self with (how, point)],
	    selector (void of (id, id) in Document
			   by (SelectionTool, MouseEvent)));

  if (!point || !selected)
    return [super adjustSelection (how, point)
		  of doc incremental incremental_p];

  boolean psel = [point selected];
  float sx, sy, sw, sh;

  if (psel && incremental_p)
    {
      (sx, sy, sw, sh) = [point decorationBBox [self neighbours point]];
      [point set_selected FALSE];
      num_selected--;
    }
  else if (!psel)
    {
      if (!incremental_p && num_selected > 0)
	(sx, sy, sw, sh) = [self deselectPoints];
      [point set_selected TRUE];
      (sx, sy, sw, sh) = [tag unionRect (sx, sy, sw, sh)
			      with [point decorationBBox
					  [self neighbours point]]];
      num_selected++;
    }
  else if (selected)
    return;

  [doc display (sx, sy, sw, sh)];
}

<doc> Add the {path} plugit to {edit}.  Nothing is needed to {display},
    since the {Shape} class is built-in.  </doc>
void
  collectPlugIts (MutableKeyed, MutableKeyed) (display, edit)
{
  [edit add [PlugIt named "path"]];
}

<doc> Set selected on the object returned by {super}.  A shape must have
    its {selected} be {TRUE} so it will display the (necessary)
    decoration.  </doc>
id (copy)
  copyForModification
{
  = [super copyForModification];
  [copy set_selected TRUE];
}

<doc> Return the normal bbox, or, in case this shape is selected, the one
    including the bbox of the points.  </doc>
(float, float, float, float) (x, y, w, h)
  decoratedBBox
{
  (x, y, w, h) = [self bbox];
  if (selected)
    {
      int i, n = [points length];
      if (n > 0)
	{
	  Point previous, current, next = points[0];

	  for (i = 0; i < n; i++)
	    {
	      (previous, current) = (current, next);
	      next = i < n - 1 ? points[i + 1] : nil;

	      (x, y, w, h)
		= [tag unionRect (x, y, w, h)
		       with [current decoratedBBox (previous, next)]];
	    }
	}
    }
}

// The doc explains why the level argument to deepen is partially bogus...
// Mon Apr 13 14:43:07 1998, tiggr@gerbil.org
<doc> If {level == 0} nothing is done.  Otherwise, the array of points is
    deepened.  This array is always deepened by 2 since it is of no
    business of the caller what number of indirection levels we use to
    store our valuable information.  </doc>
id (self)
    deepen int level
  mutably: boolean mutable_p = NO
{
  if (level != 0)
    points = [[points copy] deepen -1];
}

<doc> Unselect all selected points.

    See also {(float, float, float, float) deselectPoints}.  </doc>
void
  deselectPoints
{
  if (!num_selected)
    return;
  int i, n = [points length];
  for (i = 0; i < n; i++)
    {
      Point p = points[i];
      if ([p selected])
	{
	  [p set_selected FALSE];
	  if (!--num_selected)
	    break;
	}
    }
}

<doc> Unselect all selected points, and return the bounding box to be
    redrawn.

    See also {void deselectPoints}.  </doc>
(float, float, float, float) (x, y, w, h)
  deselectPoints
{
  if (!num_selected)
    return;
  int i, n = [points length];
  for (i = 0; i < n; i++)
    {
      Point p = points[i];
      if ([p selected])
	{
	  (x, y, w, h) = [tag unionRect (x, y, w, h)
			      with [p decorationBBox [self neighbours i]]];
	  [p set_selected FALSE];
	  if (!--num_selected)
	    break;
	}
    }
}


<doc> Tell the {points} to draw themselves.  </doc>
void
  drawSelection (float, float, float, float) (x, y, w, h)
	     in Document doc
{
  int i, n = [points length];

  if (n > 0)
    {
      Point previous, current, next = points[0];

      for (i = 0; i < n; i++)
	{
	  (previous, current) = (current, next);
	  next = i < n - 1 ? points[i + 1] : nil;

	  [current drawSelection (x, y, w, h) in doc with (previous, next)];
	}
    }
}

void
  drawSelf (float, float, float, float) (x, y, w, h)
	in Document doc
{
  Display d = [doc currentDisplay];
  [d setColor [Color black]];
  [self setupCurrentPath d];
  [d stroke];
}

// Check that tomc issues a warning if this method is declared to redefine
// something, and it isn't actually redefining something.  This is to
// catch errors when the same method of super is changed (and this method
// no longer actually redefines an existing method).
// Mon Apr 13 01:13:32 1998, tiggr@gerbil.org
<doc> See if any of the points is hit.  If it is, the {object} returned is
    {self}, and the {point} is the actual point being hit.  </doc>
redefine (Object, int, Point, float, int)
	 (object, action, point, distance, priority)
  hitTest (float, float) (x, y)
  partial boolean partial_p
{
  int i, a, p, hit = -1, n = [points length];
  float bx, by, bw, bh, px, py, d;
  boolean hp;

  if (n > 0)
    {
      Point previous, current, next = points[0];

      for (i = 0; i < n; i++)
	{
	  (previous, current) = (current, next);
	  next = i < n - 1 ? points[i + 1] : nil;

	  (hp, a, d, p) = [current hitTest (x, y) with (previous, next)];
	  if (hp && (!object || d < distance))
	    = (self, a, current, d, p);
	}
    }
}

<doc> Move a control point (indicated by {how}) of the point at {index} in
    {points}, by {(dx, dy)} from the {original}.  </doc>
protected (float, float, float, float) (x, y, w, h)
	  move (float, float) (dx, dy)
  controlPoint int how
	    at int index
      original Point original
{
  Point point = points[index], previous, next;
  (previous, next) = [self neighbours index];

  /* If this is the first time we're invoked, none of the point will have
     decoration, since we have just been created through a copy.  In that
     case, make the POINT selected.  */
  if (!num_selected)
    {
      [point set_selected TRUE];
      num_selected++;
      (x, y, w, h) = [self decoratedBBox];
    }

  (x, y, w, h) = [tag unionRect (x, y, w, h)
		      with [point decoratedBBox (previous, next)]];
  [point move (dx, dy) controlPoint how from original];
  (x, y, w, h) = [tag unionRect (x, y, w, h)
		      with [point decoratedBBox (previous, next)]];
  if (previous != nil)
    (x, y, w, h) = [tag unionRect (x, y, w, h)
		      with [previous controlDecorationBBox FALSE]];
  if (next != nil)
    (x, y, w, h) = [tag unionRect (x, y, w, h)
		      with [next controlDecorationBBox TRUE]];
}

<doc> Selector suffix for a curried invocation of the method
    {moveControlPoint of in by}.  </doc>
deferred void
  of (id, id) (original, replacement)
  in Document doc
  by (SelectionTool, MouseEvent) (tool, event);

<doc> Invoked by the {SelectionTool} (through an {Invocation}) to move a
    control point.  {relative} is one of the {SHAPE_POINT_*} constants,
    indicating whether the incoming or outgoing curve control point is
    being moved.  </doc>
void
  moveControlPoint (int, Point) (relative, point)
		of (id, id) (original, replacement)
		in Document doc
		by (SelectionTool, MouseEvent) (tool, event)
pre
  self == original
{
  int i = [points indexOfIdentical point];
  float dx, dy, nx, ny;
  (dx, dy) = [tool mouseDownLocation];
  (nx, ny) = [tool mouseDragLocation];
  (dx, dy) = (nx - dx, ny - dy);

  [doc display [replacement move (dx, dy)
			    controlPoint relative
			    at i original point]];
}

<doc> Move all or the selected points and recompute the bounding box.
    </doc>
void
  moveSelectedRelative (float, float) (dx, dy)
	    orAbsolute (float, float) (x, y)
		  from id original
{
  int i, n = [points length], left = !num_selected ? n : num_selected;

  for (i = 0; i < n; i++)
    {
      Point p = points[i];

      if (!num_selected || [p selected])
	{
	  [p moveBy (dx, dy)];
	  if (!--left)
	    break;
	}
    }

  if (!num_selected)
    {
      /* We have just moved without reshaping.  */
      [super moveSelectedRelative (dx, dy) orAbsolute (x, y) from original];
    }
  else
    [self recomputeBBox];
}


<doc> Return the neighbouring points of the point {p}.

    See also {(Point, Point) neighbours int index}.  </doc>
(Point, Point) (previous, next)
  neighbours Point p
{
  int i = [points indexOfIdentical p];

  if (i > 0)
    previous = points[i - 1];
  if (i < [points length] - 1)
    next = points[i + 1];
}

<doc> Return the neighbouring points of the point {p}.

    See also {(Point, Point) neighbours Point p}.  </doc>
(Point, Point) (previous, next)
  neighbours int index
{
  if (index > 0)
    previous = points[index - 1];
  if (index < [points length] - 1)
    next = points[index + 1];
}

<doc> Return {TRUE}.  </doc>
boolean
  partialRubberSuffices
{
  = TRUE;
}

<doc> Recompute the (undecorated) bounding box of the receiving shape, as
    stored in the {(b_x, b_y, b_w, b_h)} ivars.  </doc>
protected void
  recomputeBBox
{
  int i, n = [points length];
  float x, y;

  b_w = b_h = 1.0;
  (b_x, b_y) = [points[0] location];
  for (i = 0; i < n; i++)
    {
      Point p = points[i];
      (b_x, b_y, b_w, b_h) = [tag unionRect (b_x, b_y, b_w, b_h) with [p bbox]];
    }
}

<doc> Remove the point from index {i} and recompute the bounding box.  The
    index can be both the length - 1 or {SHAPE_POINT_LAST} to remove the
    last point.  Return the point that was removed.  </doc>
Point (point)
  removePoint int i
pre
  SHAPE_POINT_LAST <= i && i < [points length]
post
  (point != nil
   && [points length] == old ([points length]) - 1)
{
  if (i == SHAPE_POINT_LAST)
    i = [points length] - 1;
  = [points removeAt i];

  if ([point selected])
    num_selected--;

  [self recomputeBBox];
}

<doc> If only the start or end point is selected, return that point.  If
    the shape only has one point, return it as the {previous} point.  In
    both cases, the point is deselected before being returned.  </doc>
(Shape, Point, Point) (shape, previous, next)
  selectedShapeSinglePoint
{
  if (num_selected == 1)
    {
      int n = [points length];
      Point p = points[n - 1];
      if ([p selected])
	{
	  = (self, p, nil);
	  num_selected--;
	  [p set_selected FALSE];
	}
      else if (n > 1)
	{
	  p = points[0];
	  if ([p selected])
	    {
	      = (self, nil, p);
	      num_selected--;
	      [p set_selected FALSE];
	    }
	}
    }
}

<doc> Make the {point} be selected, and return the bounding box to be
    redrawn to make its decoration visible.  </doc>
(float, float, float, float)
  selectPoint Point point
pre
  [points indexOfIdentical point] >= 0
{
  if (![point selected])
    {
      [point set_selected TRUE];
      num_selected++;
      = [point decorationBBox [self neighbours point]];
    }
}

<doc> Apply rubber banding selection to the points of the receiving object
    for the given rectangle.  </doc>
(float, float, float, float) (rx, ry, rw, rh)
  selectRubber (float, float, float, float) (x, y, w, h)
	    in Document doc
{
  if (!selected && [tag rect (x, y, w, h) intersects [self bbox]])
    {
      int i, n = [points length];
      float px, py;

      for (i = 0; i < n; i++)
	{
	  Point p = points[i];
	  (px, py) = [p location];

	  if (px >= x && px < x + w && py >= y && py < y + h)
	    {
	      if (!selected)
		[doc addToSelection self];
	      [p set_selected TRUE];
	      num_selected++;
	    }
	}

      if (selected)
	= [self decoratedBBox];
    }
}

<doc> If going from unselected to selected, first unselect all points.
    </doc>
void
  set_selected boolean new_value
{
  if (!selected && new_value && num_selected > 0)
    [self deselectPoints];
  [super set_selected new_value];
}

<doc> Set the current path on the display {d} to our path.  This can be
    followed by for instance a stroke to produce the desired effect.
    </doc>
void
  setupCurrentPath Display d
{
  int i, n = [points length];

  [d newpath];

  if (n > 0)
    {
      Point previous, current, next = points[0];

      for (i = 0; i < n; i++)
	{
	  (previous, current) = (current, next);
	  next = i < n - 1 ? points[i + 1] : nil;

	  [current drawSelf d with (previous, next)];
	}
    }
}

<doc> Emit the location.  </doc>
OutputStream
  writeFields OutputStream s
{
  = [[super (Object) writeFields s] print (" points=", points)];
}

/******************** Archiving ********************/
<doc><h4>Archiving</h4></doc>

void
  encodeUsingCoder Encoder coder
{
  if (![coder hasBeenCodedFor [Shape self]])
    {
      [super encodeUsingCoder coder];
      [coder encode points];
      [coder encode closed];
    }
}

void
  initWithCoder Decoder coder
{
  class (id) c = [Shape self];

  if (![coder hasBeenCodedFor c])
    {
      int version = [coder versionOfClass c];

      [super initWithCoder coder];
      points = [coder decode];
      if (version > 0)
	closed = [coder decode];
    }
}

end;
