<copyright> It SimpleObject class.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of It.  It is distributed under the terms of the
    It License, a copy of which can be found in the It distribution; see
    the file LICENSE.

    <id>$Id: SimpleObject.t,v 1.2 1999/04/14 22:27:34 tiggr Exp $</id>
    </copyright>

<doc> A {SimpleObject} is an {Object} defined by two points, one
    bottom-left, and the other top-right.  </doc>
implementation class
SimpleObject: Object, Math

end;

implementation instance
SimpleObject
{
  <doc> The four locations.  </doc>
  float blx, bly, tlx, tly, brx, bry, trx, try;
}

<doc> Designated initializer.  </doc>
id
  initWithBottomLeft (float, float) (bx, by)
	    topRight (float, float) (tx, ty)
		bbox (float, float, float, float) (x, y, w, h)
{
  (blx, bly, trx, try) = (bx, by, tx, ty);
  (tlx, tly, brx, bry) = (bx, ty, tx, by);
  = [super initWithBBox (x, y, w, h)];
}

<doc> If {how} indicates one of the points, start resizing the receiving
    object.  Otherwise, let {super} do its job.  </doc>
(All, selector) (target, action)
  adjustSelection (int, All) (how, extra)
	       of Document doc
      incremental boolean incremental_p
{
  if (how >= 0 && how <= 3)
    {
      = ([Invocation for selector (void resizePoint int
					of (id, id) in Document
					by (SelectionTool, MouseEvent))
		     to: self with how],
	 selector (void of (id, id) in Document
			by (SelectionTool, MouseEvent)));
      [doc clearSelection];
    }
  else
    = [super adjustSelection (how, extra) of doc incremental incremental_p];
}

<doc> Return the normal bbox, or, in case this object is selected, the one
    including the bbox of the points.  </doc>
(float, float, float, float) (x, y, w, h)
  decoratedBBox
{
  (x, y, w, h) = [self bbox];
  if (selected)
    {
      float iw, ih;
      (iw, ih) = [point_image size];

      = [tag unionRect (x, y, w, h)
	     with (blx - iw / 2.0 + 0.5, bly - iw / 2.0 + 0.5, iw, ih)];
      = [tag unionRect (x, y, w, h)
	     with (tlx - iw / 2.0 + 0.5, tly - iw / 2.0 + 0.5, iw, ih)];
      = [tag unionRect (x, y, w, h)
	     with (brx - iw / 2.0 + 0.5, bry - iw / 2.0 + 0.5, iw, ih)];
      = [tag unionRect (x, y, w, h)
	     with (trx - iw / 2.0 + 0.5, try - iw / 2.0 + 0.5, iw, ih)];
    }
}

<doc> Return the four points.  </doc>
protected (float, float, float, float, float, float, float, float)
  dissect
{
  = (blx, bly, tlx, tly, brx, bry, trx, try);
}

<doc> Draw the four points.  </doc>
void
  drawSelection (float, float, float, float) (x, y, w, h)
	     in Document doc
{
  // Intersect box...
  // Sun Apr 19 00:46:52 1998, tiggr@gerbil.org
  [isa centerImage point_image at (blx, bly)];
  [isa centerImage point_image at (tlx, tly)];
  [isa centerImage point_image at (brx, bry)];
  [isa centerImage point_image at (trx, try)];
}

<doc> See if any of the points is hit.  If it is, the {action} returned
    ranges from 0 to 3 (inclusive) to indicate which point was hit.  If we
    are not currently selected, then the points can not be hit.  </doc>
redefine (Object, int, Point, float, int)
	 (object, action, point, distance, priority)
  hitTest (float, float) (x, y)
  partial boolean partial_p
{
  action = -1;

  (object, distance, priority) = [self hitTest (x, y) point (blx, bly)];
  if (object != nil)
    action = 0;
  else
    {
      (object, distance, priority) = [self hitTest (x, y) point (tlx, tly)];
      if (object != nil)
	action = 1;
      else
	{
	  (object, distance, priority) = [self hitTest (x, y) point (brx, bry)];
	  if (object != nil)
	    action = 2;
	  else
	    {
	      (object, distance, priority)
		= [self hitTest (x, y) point (trx, try)];
	      if (object != nil)
		action = 3;
	    }
	}
    }

  if (!selected)
    action = -1;
}

<doc> Return {self} if {(x, y)} hits the point at {(px, py)}.  </doc>
protected (id, float, int) (object, distance, priority)
  hitTest (float, float) (x, y)
    point (float, float) (px, py)
{
  distance = float (sqrt ((x - px) * (x - px) + (x - py) * (y - py)));
  // Fully arbitrary...
  // Sun Apr 19 12:54:52 1998, tiggr@gerbil.org
  if (distance < 3.0)
    (object, priority) = (self, HIT_PRIO_DETAIL);
}

<doc> Move the receiving object by {(dx, dy)}.  The default implementation
    moves the points and (lets super move) the bounding box.  </doc>
void
  moveSelectedRelative (float, float) (dx, dy)
	    orAbsolute (float, float) (x, y)
		  from id original
{
  (blx, bly, tlx, tly) = (blx + dx, bly + dy, tlx + dx, tly + dy);
  (brx, bry, trx, try) = (brx + dx, bry + dy, trx + dx, try + dy);
  [super moveSelectedRelative (dx, dy) orAbsolute (x, y) from original];
}

<doc> Recompute the (undecorated) bounding box.  Default implementation
    recomputes a bounding box of 1 point around the points.  </doc>
protected void
  recomputeBBox
{
  float lx = blx - 1.0, by = bly - 1.0, rx = blx + 1.0, ty = by + 1.0;

  if (tlx - 1.0 < lx)
    lx = tlx - 1.0;
  if (brx - 1.0 < lx)
    lx = brx - 1.0;
  if (trx - 1.0 < lx)
    lx = trx - 1.0;

  if (tly - 1.0 < by)
    by = tly - 1.0;
  if (bry - 1.0 < by)
    by = bry - 1.0;
  if (try - 1.0 < by)
    by = try - 1.0;

  if (tlx + 1.0 > rx)
    rx = tlx + 1.0;
  if (brx + 1.0 > rx)
    rx = brx + 1.0;
  if (trx + 1.0 > rx)
    rx = trx + 1.0;

  if (tly + 1.0 > ty)
    ty = tly + 1.0;
  if (bry + 1.0 > ty)
    ty = bry + 1.0;
  if (try + 1.0 > ty)
    ty = try + 1.0;

  (b_x, b_y, b_w, b_h) = (lx, by, rx - lx, ty - by);
}

protected void
   resize (float, float) (dx, dy)
    point int pt
     from id original
  options (boolean, boolean) (square, center)
pre
  pt >= 0 && pt <= 3
{
  (blx, bly, tlx, tly, brx, bry, trx, try) = [original dissect];
  if (pt == 0)
    {
      (tlx, bry) = (blx, bly) = (blx + dx, bly + dy);
      if (center)
	(brx, tly) = (trx, try) = (trx - dx, try - dy);
    }
  else if (pt == 1)
    {
      (blx, try) = (tlx, tly) = (tlx + dx, tly + dy);
      if (center)
	(trx, bly) = (brx, bry) = (brx - dx, bry - dy);
    }
  else if (pt == 2)
    {
      (trx, bly) = (brx, bry) = (brx + dx, bry + dy);
      if (center)
	(blx, try) = (tlx, tly) = (tlx - dx, tly - dy);
    }
  else
    {
      (brx, tly) = (trx, try) = (trx + dx, try + dy);
      if (center)
	(tlx, bry) = (blx, bly) = (blx - dx, bly - dy);
    }

  [self recomputeBBox];
}

<doc> Selector suffix for a curried invocation of the method
    {resizePoint of in by}.  </doc>
deferred void
  of (id, id) (original, replacement)
  in Document doc
  by (SelectionTool, MouseEvent) (tool, event);

<doc> Further resize the {replacement} object, while dragging the point
    indicated by {pt}.  </doc>
void
  resizePoint int pt
	   of (id, id) (original, replacement)
	   in Document doc
	   by (SelectionTool, MouseEvent) (tool, event)
pre
  self == original
  && pt >= 0 && pt <= 3
{
  float dx, dy, nx, ny;
  (dx, dy) = [tool mouseDownLocation];
  (nx, ny) = [tool mouseDragLocation];
  (dx, dy) = (nx - dx, ny - dy);

  float x, y, w, h;
  (x, y, w, h) = [replacement decoratedBBox];

  int mods = [event modifiers];
  // ratio-preserving...
  // Sun Apr 19 14:51:41 1998, tiggr@gerbil.org
  boolean square = !!(mods & EVENT_MOD_SHIFT);
  boolean center = !!(mods & EVENT_MOD_CONTROL);

  [replacement resize (dx, dy) point pt from original options (square, center)];
  [doc display [tag unionRect (x, y, w, h)
		    with [replacement decoratedBBox]]];
}

<doc> Update the two points and recompute the bounding box.  </doc>
void
  setBottomleft (float, float) (lx, by)
       topRight (float, float) (rx, ty)
{
  (blx, bly, trx, try) = (lx, by, rx, ty);
  (tlx, tly, brx, bry) = (lx, ty, rx, by);
  [self recomputeBBox];
}

/******************** Archiving ********************/
<doc><h4>Archiving</h4></doc>

void
  encodeUsingCoder Encoder coder
{
  if (![coder hasBeenCodedFor [SimpleObject self]])
    {
      [super encodeUsingCoder coder];
      [coder encode blx];
      [coder encode bly];
      [coder encode tlx];
      [coder encode tly];
      [coder encode brx];
      [coder encode bry];
      [coder encode trx];
      [coder encode try];
    }
}

void
  initWithCoder Decoder coder
{
  class (id) c = [SimpleObject self];
  if (![coder hasBeenCodedFor c])
    {
      [super initWithCoder coder];
      blx = [coder decode];
      bly = [coder decode];
      tlx = [coder decode];
      tly = [coder decode];
      brx = [coder decode];
      bry = [coder decode];
      trx = [coder decode];
      try = [coder decode];
    }
}

end;
