<copyright> It SimpleObjectTool class.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of It.  It is distributed under the terms of the
    It License, a copy of which can be found in the It distribution; see
    the file LICENSE.

    <id>$Id: SimpleObjectTool.t,v 1.1 1998/04/19 14:59:58 tiggr Exp $</id>
    </copyright>

<doc> An SimpleObjectTool is a tool to create objects whose appearance
    only depends on their bounding box, such as rectangles and ellipses.
    This tool can not be used as such; it must be subclassed.  </doc>
implementation class
SimpleObjectTool: Tool

end;

implementation instance
SimpleObjectTool
{
  <doc> The most recent mouse down location, in the coordinate system of
      the document.  </doc>
  float down_x, down_y;

  <doc> Likewise, the most recent mouse drag location.  </doc>
  float drag_x, drag_y;

  <doc> The object being created.  </doc>
  SimpleObject object;
}

<doc> Allocate and return a new instance of the {objectClass}.  </doc>
SimpleObject
  newObject (float, float, float, float) (lx, by, rx, ty)
{
  // This is incorrect
  // Sat Apr 18 23:58:23 1998, tiggr@gerbil.org
  = [[[self objectClass] alloc]
      initWithBottomLeft (lx, by) topRight (rx, ty)
      bbox (lx - 1.0, by - 1.0, rx - lx + 2.0, ty - by + 2.0)];
}

<doc> Return the class of the objects created by this tool.  </doc>
deferred class (SimpleObject)
  objectClass;

<doc> Tell the {object} of the new locations of the points.  </doc>
void
  updateObject (float, float, float, float) (lx, by, rx, ty)
{
  [object setBottomleft (lx, by) topRight (rx, ty)];
}

/******************** Tool ********************/

<doc> Draw the {object}.  </doc>
void
  drawActivity (float, float, float, float) (x, y, w, h)
	    in Document doc
{
  if (object != nil)
    [object drawSelf (x, y, w, h) in doc];
}

boolean
  mouseDown MouseEvent e
	 in Document doc
{
  [doc clearSelection];
  (down_x, down_y) = (drag_x, drag_y) = [e location];
  = TRUE;
}

void
  mouseDragged MouseEvent e
	    in Document doc
{
  float x, y;
  (x, y) = [e location];

  int mods = [e modifiers];
  boolean square = !!(mods & EVENT_MOD_SHIFT);
  if (square)
    if (x > down_x)
      if (y > down_y)			/* Upper right.  */
	if (x - down_x > y - down_y)
	  y = down_y + x - down_x;
	else
	  x = down_x + y - down_y;
      else if (x - down_x > down_y - y)	/* Lower right.  */
	y = down_y - x + down_x;
      else
	x = down_x + down_y - y;
    else if (y > down_y)		/* Upper left.  */
      if (down_x - x > y - down_y)
	y = down_y + down_x - x;
      else
	x = down_x - y + down_y;
    else if (down_x - x > down_y - y)	/* Lower left.  */
      y = down_y + x - down_x;
    else
      x = down_x - down_y + y;
      
  if (x == drag_x && y == drag_y)
    return;
  (drag_x, drag_y) = (x, y);
  float w = drag_x - down_x, h = drag_y - down_y;

  boolean center_down = !!(mods & EVENT_MOD_CONTROL);
  if (center_down)
    (x, y, w, h) = (down_x - (drag_x - down_x), down_y - (drag_y - down_y),
		    2.0 * w, 2.0 * h);
  else
    (x, y) = (down_x, down_y);

  (x, y, w, h) = fix_rectangle (x, y, w, h);

  float rdx, rdy, rdw, rdh;
  if (!object)
    object = [self newObject (x, y, x + w, y + h)];
  else
    {
      (rdx, rdy, rdw, rdh) = [object decoratedBBox];
      [self updateObject (x, y, x + w, y + h)];
    }
  [doc display [tag unionRect (rdx, rdy, rdw, rdh)
		    with [object decoratedBBox]]];
}

void
  mouseUp MouseEvent e
       in Document doc
{
  if (down_x != drag_x && down_y != drag_y)
    [doc addObject object];
  object = nil;
}

end;
