<copyright> It ToolKit class.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of It.  It is distributed under the terms of the
    It License, a copy of which can be found in the It distribution; see
    the file LICENSE.

    <id>$Id: ToolKit.t,v 1.10 1999/04/14 22:27:35 tiggr Exp $</id>
    </copyright>

implementation class
ToolKit: it, MatrixDelegate
{
  <doc> The (for now) shared tool kit.  </doc>
  static instance (id) shared;
}

<doc> Return the shared and only ToolKit instance.  </doc>
instance (id)
  shared
{
  if (!shared)
    shared = [self new];
  = shared;
}

end;

implementation instance
ToolKit
{
  <doc> The window displaying this tool kit.  </doc>
  Panel window;

  <doc> The matrix of tool buttons.  </doc>
  Matrix buttons;

  <doc> The tools, in the same order as the cells of the buttons
      (row-major).  </doc>
  MutableIndexed tools;

  <doc> The selected tool.  </doc>
  public Tool selected_tool;
}

id
  init
{
  Screen s = [system_display defaultScreen];

  window = [s panelWithContentRect (50.0, 600.0, 66.0, 200.0)
	      styleMask WINDOW_STYLE_TITLED backing BACKING_STORE_RETAINED];
  [window setTitle "ToolkIt"];

  ButtonCell ac = [[ButtonCell alloc] initTextCell ""];
  [ac setImagePosition BUTTONCELL_IMAGE_ONLY];
  int how = (CELL_MASK_CHANGE_BACKGROUND | CELL_MASK_CHANGE_GRAY
	     | CELL_MASK_PUSH | CELL_MASK_CONTENTS);
  [ac setHighlightsBy how];
  [ac setShowsStateBy how];
  [ac set_bordered YES];
  [ac set_image [Image named "tool-empty"]];
  [ac set_alternate_image [Image named "tool-empty-h"]];
  buttons = [[Matrix alloc] initFrame [[window content_view] frame]
			    mode MATRIX_MODE_RADIO prototype ac
			    dimensions (1, 2)];
  [buttons setAction selector (void selectionBy Matrix)];
  [buttons setIntercellSpacing (4.0, 4.0)];
  [buttons set_auto_size VIEW_SIZE_MINX | VIEW_SIZE_MAXX | VIEW_SIZE_MINY];
  [buttons setTarget self];
  [buttons sizeToCells];

  float x, y, w, h, mw, mh;
  (, , mw, mh) = [buttons frame];
  View cv = [window content_view];
  (x, y, w, h) = [cv frame];
  [buttons moveTo (x + (w - mw) / 2.0, y + h - mh - 2.0)];
  [cv addSubview buttons];

  tools = [MutableObjectArray new];
  [self newTool [SelectionTool self]];

  [window orderFront self];

  = [super init];
}

void
  newTool class (Tool) tool_class
{
  int n_t = [tools length];
  Indexed cells = [buttons cells];
  int n_c = [cells length];

  if (n_t == n_c)
    {
      [buttons addRow];
      [buttons sizeToFit];

      float x, y, w, h, oh;

      (, , w, h) = [buttons frame];
      (x, y, , oh) = [window frame];
      [window setFrame (x, y + oh - h, w + 5.0, h + 4.0)];
      /* Having resized the window will have moved the buttons
	 (due to auto sizing).  */
      [buttons moveTo (2.0, 3.0)];
    }

  ButtonCell ac = [buttons cells][n_t];

  Tool tool = [tool_class withCell ac];

  /* Using an invocation that is curried on the first argument (the tool)
     as the target means we do not have to do any searching to find the
     tool given the matrix' selection.  */
  [ac setTarget [Invocation for selector (void tool Tool
					       selectionBy Matrix)
			    to: self with tool]];

  [tools add tool];
  if ([buttons selected_cell] == ac)
    selected_tool = tool;
}

/******************** internal methods ********************/

<doc> Plain method for when the user selects a tool (button) which
    actually is not a tool yet.  This would be the bottom-right button
    when we have an odd number of tools.  </doc>
void
  selectionBy Matrix sender
{
}

<doc> Make the {tool} the {selected_tool}.  </doc>
void
	 tool Tool tool
  selectionBy Matrix sender
{
  selected_tool = tool;
}

end;
