<copyright> It:grid GridLayer class.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of It.  It is distributed under the terms of the
    It License, a copy of which can be found in the It distribution; see
    the file LICENSE.

    <id>$Id: GridLayer.t,v 1.8 1998/04/10 20:48:09 tiggr Exp $</id>
    </copyright>

implementation class
GridLayer: Layer
{
  <doc> Half the size of the major crosshairs.  </doc>
  const MAJOR_DISP = 2.0;
}

end;

implementation instance
GridLayer
{
  <doc> The major grid distance.  </doc>
  float major;

  <doc> The minor grid distance.  </doc>
  float minor;
}

<doc> Designated initializer.  </doc>
id
  initWithDocument Document d
{
  (major, minor) = (50.0, 5.0);
  = [super initWithDocument d];
}

// Technically speaking, we do not need a grid to display a document.
// However, until we have some kind of archiving proxy, the Grid class
// needs to be included, since it is referenced in the archived document.
// Sun Mar 22 16:52:21 1998, tiggr@gerbil.org
<doc> Add the {grid} to the {display} set.  </doc>
void
  collectPlugIts (MutableKeyed, MutableKeyed) (display, edit)
{
  [display add [grid shared]];
}

<doc> Draw the grid.  (But not if the document is busy printing.)  </doc>
void
  drawSelf (float, float, float, float) (x, y, w, h)
	in Document doc
{
  if ([doc printing])
    return;

  float bw, bh, mx, my, gx, gy;

  (, , bw, bh) = [doc bounds];

  Display d = [doc currentDisplay];
  [d setColor [Color light]];
  for (mx = 0.0; mx < bw; mx += major)
    for (my = 0.0; my < bh; my += major)
      if (mx < x + w + MAJOR_DISP && mx + major >= x - MAJOR_DISP
	  && my < y + h + MAJOR_DISP && my + major >= y - MAJOR_DISP)
	  {
	    if (mx - MAJOR_DISP < x + w && mx + MAJOR_DISP >= x
		&& my - MAJOR_DISP < y + h && my + MAJOR_DISP >= y)
	      {
		[d drawLine (mx - MAJOR_DISP, my, mx + MAJOR_DISP, my)];
		[d drawLine (mx, my - MAJOR_DISP, mx, my + MAJOR_DISP)];
	      }

	    for (gx = minor; gx < major; gx += minor)
	      if (mx + gx >= x && mx + gx < x + w
		  && my >= y && my < y + h)
		[d drawLine (mx + gx, my, mx + gx, my)];

	    for (gy = minor; gy < major; gy += minor)
	      if (mx >= x && mx < x + w
		  && my + gy >= y && my + gy < y + h)
		[d drawLine (mx, my + gy, mx, my + gy)];
	  }
}

(boolean, float, float) (locked, new_x, new_y)
  lockEvent MouseEvent e
   location (float, float) (x, y)
{
  int xx = int ((x + minor / 2.0) / minor);
  int yy = int ((y + minor / 2.0) / minor);

  = (TRUE, float (xx) * minor, float (yy) * minor);
}

/******************** Archiving ********************/
<doc><h4>Archiving</h4></doc>

void
  encodeUsingCoder Encoder coder
{
  if (![coder hasBeenCodedFor [GridLayer self]])
    {
      [super encodeUsingCoder coder];
      [coder encode major];
      [coder encode minor];
    }
}

void
  initWithCoder Decoder coder
{
  if (![coder hasBeenCodedFor [GridLayer self]])
    {
      [super initWithCoder coder];
      major = [coder decode];
      minor = [coder decode];
    }
}

end;
