<copyright> It PathTool class.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of It.  It is distributed under the terms of the
    It License, a copy of which can be found in the It distribution; see
    the file LICENSE.

    <id>$Id: PathTool.t,v 1.5 1998/04/14 11:48:48 tiggr Exp $</id>
    </copyright>

/******************** path ********************/

implementation class
path: PlugIt

end;

implementation instance
path

<doc> Designated initializer.  </doc>
id
  init
{
  = [super init];
  [[ToolKit shared] newTool [PathTool self]];
}

String
  plugItName
{
  = "path";
}

int
  plugItVersion
{
  = 1;
}

end;

/******************** PathTool ********************/

implementation class
PathTool: Tool

<doc> Return {"tool-path"} and {"tool-path-h"}.  </doc>
(String, String) (normal, alternate)
  cell_images
{
  = ("tool-path", "tool-path-h");
}

end;

implementation instance
PathTool
{
  <doc> The shape to which we will be adding the new point.  </doc>
  Shape shape;

  <doc> The point preceding and following the new point.  </doc>
  Point previous, next;

  <doc> The point that we are manipulating.  </doc>
  Point point;
}

void
  drawActivity (float, float, float, float) (x, y, w, h)
	    in Document doc
{
  [point drawSelection (x, y, w, h) in doc with (previous, next)];

  // Maybe this point should already be part of the SHAPE, so we do not
  // need to draw things here...
  // Mon Mar 16 01:19:39 1998, tiggr@gerbil.org
  if (previous != nil || next != nil)
    {
      Display d = [doc currentDisplay];
      float x1, y1, x2, y2, x3, y3;

      [d newpath];
      if (previous != nil)
	{
	  [d moveto [previous location]];
	  (x1, y1) = [previous locationOutgoing];
	  (x2, y2) = [point locationIncoming];
	  (x3, y3) = [point location];
	}
      else
	{
	  [d moveto [point location]];
	  (x1, y1) = [point locationOutgoing];
	  (x2, y2) = [next locationIncoming];
	  (x3, y3) = [next location];
	}
      [d curveto (x1, y1, x2, y2, x3, y3)];
      [d setColor [Color black]];
      [d stroke];

      if (previous != nil)
	[previous drawSelection (x, y, w, h) in doc with (nil, point)];
      if (next != nil)
	[next drawSelection (x, y, w, h) in doc with (point, nil)];
    }
}

boolean
  mouseDown MouseEvent e
	 in Document doc
{
  (shape, previous, next) = [doc selectedShapeSinglePoint];

  point = [[Point alloc] init nil position [e location]];
  [point set_selected TRUE];

  if (shape != nil)
    [doc display [shape decoratedBBox]];
  
  = TRUE;
}

void
  mouseDragged MouseEvent e
	    in Document doc
{
  float sx, sy, sw, sh;
  float px, py, mx, my;

  (sx, sy, sw, sh) = [point decoratedBBox (previous, next)];
  (px, py) = [point location];
  (mx, my) = [e location];

  if (previous != nil)
    (sx, sy, sw, sh) = [tag unionRect (sx, sy, sw, sh)
			    with [previous decoratedBBox (nil, point)]];
  else if (next != nil)
    (sx, sy, sw, sh) = [tag unionRect (sx, sy, sw, sh)
			    with [next decoratedBBox (point, nil)]];

  [point set_is_curve];
  [point set_control_points (px - (mx - px), py - (my - py), mx, my)];
  [doc display [tag unionRect (sx, sy, sw, sh)
		    with [point decoratedBBox (previous, next)]]];
}

void
  mouseUp MouseEvent e
       in Document doc
{
  if (!shape)
    {
      Shape s = [[Shape alloc] initWithPoint point];
      [doc addObject s];
      [doc display [s selectPoint point]];
    }
  else
    {
      Modification mod
	= [[PointAddModification alloc]
	    initWithPoint point inShape shape
	    atIndex (!previous ? SHAPE_POINT_FIRST : SHAPE_POINT_LAST)];
      [doc addModification mod];
      [doc display [mod redo doc]];
    }

  shape = nil;
  point = previous = next = nil;
}

end;
